﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Bentley.Interop.MicroStationDGN;
using System.Windows;
using System.Windows.Forms;

namespace ISY.CAD.Noise
{
    class IndustryNoiseMgr : ILocateCommandEvents
    {
        private static string typeOfOutline = string.Empty;
        private IndustryNoiseForm noiseForm;

        public IndustryNoiseMgr()
        {
            noiseForm = new IndustryNoiseForm();
        }

        internal static void StartCommand()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            IndustryNoiseMgr command = new IndustryNoiseMgr();
            App.CommandState.StartLocate(command);
            App.CommandState.StartDynamics();
        }


        private void AddElement(Element Element)
        {
            List<Point3d> line = new List<Point3d>();
            if ((Element.Type == MsdElementType.LineString) || (Element.Type == MsdElementType.Line))
            {
                LineElement el = Element.AsLineElement();
                foreach (Point3d p in el.GetVertices())
                    if (!line.Contains(p))
                        line.Add(p);
                typeOfOutline = "Line";
            }
            else if (Element.Type == MsdElementType.CellHeader)
            {
                CellElement el = Element.AsCellElement();
                ElementEnumerator ee = el.GetSubElements();
                while (ee.MoveNext())
                {
                    ComplexShapeElement shape;
                    shape = ee.Current.AsComplexShapeElement();
                    ElementEnumerator ee2 = shape.GetSubElements();
                    while (ee2.MoveNext())
                    {
                        LineElement e;
                        e = ee2.Current.AsLineElement();
                        foreach (Point3d p in e.GetVertices())
                            line.Add(p);
                        typeOfOutline = "Cell";
                    }
                }
            }
            else if (Element.Type == MsdElementType.Shape)
            {
                ShapeElement el = Element.AsShapeElement();
                foreach (Point3d p in el.GetVertices())
                    line.Add(p);
                typeOfOutline = "Shape";
            }
            AddIndustryNoiseSource(line);
        }

        private void AddIndustryNoiseSource(List<Point3d> line)
        {
            double[,] points = RunCalc.convertPoints(line.ToArray());
            float amount = noiseForm.GetAmount();
            string activityType = noiseForm.GetNoiseSource();
            string distributionType = noiseForm.getDistributionType();
            string geometry = noiseForm.getGeometry();

            // checking if input is valid:
            if (amount > 24 || amount < 0)
            {
                System.Windows.Forms.MessageBox.Show("Error! Amount needs to be between 0 and 24 hours", "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            // test if point is only one point
            if (geometry == "point" && points.GetLength(0) != 1)
            {
                System.Windows.Forms.MessageBox.Show("Error! A point needs to be defined by only a point",
                    "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }
           
            // test that lines are at least two points
            if (geometry == "line" && points.GetLength(0) < 2)
            {
                System.Windows.Forms.MessageBox.Show("Error! A line needs to be defined by at least two points",
                    "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            // test that areas and volumes are at least 4 points
            if (geometry == "area" && points.GetLength(0) < 4)
            {
                System.Windows.Forms.MessageBox.Show("Error! An area needs to be defined by at least four points",
                    "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            if (geometry == "volume" && points.GetLength(0) < 4)
            {
                System.Windows.Forms.MessageBox.Show("Error! A volume needs to be defined by at least four points",
                    "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            if (geometry == "area" || geometry == "volume")
            {
                double zHeight = points[0, 2];
                bool isSameHeight = true;
                for (int i = 1; i < points.GetLength(0); i++)
                    if (zHeight != points[i, 2])
                        isSameHeight = false;
                if (!isSameHeight)
                    System.Windows.Forms.MessageBox.Show("Warning! The points for the noise source have different heights, only the height of the first point will be used.");
            }

            // Checking if there are an odd number of points for a line:
            // if there are, add one point as SoundKernel gives
            // errors for lines with an odd number of points
            // This should be fixed in SoundKernel!
            if (geometry == "line")
            {
                int res = 0;
                Math.DivRem(line.Count, 2, out res);
                if (res != 0)
                {
                    line.Add(line[line.Count - 1]);
                    points = RunCalc.convertPoints(line.ToArray());
                }
            }




            double[] frequencySpectra = NoiseSourceSpecForm.freqSpectra;

            if (activityType == "Crusher")
                frequencySpectra = populateCrusherSourceSpec();
            else if (activityType == "Drilling")
                frequencySpectra = populateDrillingSourceSpec();
            else if (activityType == "Dumper")
                frequencySpectra = populateDumperSourceSpec();

            if (frequencySpectra == null)
            {
                System.Windows.Forms.MessageBox.Show("Error! A noise source has to be selected or a user defined noise source has to be defined.",
                    "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            double[] timeDistribution = new double[10];
            if (distributionType == "General")
                timeDistribution = null;
            else if (distributionType == "Work")
                timeDistribution = null;
            else if (distributionType == "UserDefined")
                timeDistribution = NoiseTimeDistributionForm.distribution;

            NoiseComputation.NoiseSource newSource;
            if (timeDistribution == null)
                newSource = new NoiseComputation.NoiseSource(points, activityType, distributionType, geometry, amount, frequencySpectra);
            else 
                newSource = new NoiseComputation.NoiseSource(points, activityType, distributionType, geometry, amount, frequencySpectra, timeDistribution);
        

            System.Windows.Forms.MessageBox.Show("A industry noise source with " + line.Count + " vertice(s) created");

            RunCalc.AddNoiseSource(newSource);
        }

        private double[] populateCrusherSourceSpec()
        {
            double[] octaveSpectra = new double[9];
            octaveSpectra[0] = -200;
            octaveSpectra[1] = 52.1;
            octaveSpectra[2] = 76.4;
            octaveSpectra[3] = 93.1;
            octaveSpectra[4] = 104;
            octaveSpectra[5] = 107.6;
            octaveSpectra[6] = 106.7;
            octaveSpectra[7] = 99.8;
            octaveSpectra[8] = 88;
            							

            return octaveSpectra;
        }

        private double[] populateDrillingSourceSpec()
        {
            double[] octaveSpectra = new double[9];
            octaveSpectra[0] = -200;
            octaveSpectra[1] = 106.5515169;
            octaveSpectra[2] = 115.6515169;
            octaveSpectra[3] = 106.9181835;
            octaveSpectra[4] = 112.9181835;
            octaveSpectra[5] = 112.2515169;
            octaveSpectra[6] = 115.1848502;
            octaveSpectra[7] = 115.3181835;
            octaveSpectra[8] = 115.9848502;

            return octaveSpectra;
        }

        private double[] populateDumperSourceSpec()
        {
            double[] octaveSpectra = new double[9];
            octaveSpectra[0] = -200;
            octaveSpectra[1] = 75.6;
            octaveSpectra[2] = 92.8;
            octaveSpectra[3] = 99.4;
            octaveSpectra[4] = 100.8;
            octaveSpectra[5] = 101.6;
            octaveSpectra[6] = 100.3;
            octaveSpectra[7] = 94.1;
            octaveSpectra[8] = 83.3;

            return octaveSpectra;
        }

        #region ILocateCommandEvents Members


        public void Accept(Element Element, ref Point3d Point, Bentley.Interop.MicroStationDGN.View View)
        {
            AddElement(Element);
        }

        public void Cleanup()
        {
            noiseForm.DetachFromMicroStation();
            if (IndustryNoiseForm.distributionForm != null)
                IndustryNoiseForm.distributionForm.Close();
        }

        public void Dynamics(ref Point3d Point, Bentley.Interop.MicroStationDGN.View View, MsdDrawingMode DrawMode)
        {
            
        }

        public void LocateFailed()
        {
            
        }

        public void LocateFilter(Element Element, ref Point3d Point, ref bool Accepted)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            if (Element.Type == MsdElementType.LineString || Element.Type == MsdElementType.Line)
            {
                App.ShowPrompt("Add line");
                Accepted = true;
            }
            else if (Element.Type == MsdElementType.CellHeader)
            {
                App.ShowPrompt("Add Cell");
                Accepted = true;
            }
            else if (Element.Type == MsdElementType.Shape)
            {
                App.ShowPrompt("Add Shape");
                Accepted = true;
            }
            else if (Element.Type == MsdElementType.PointString)
            {
                App.ShowPrompt("Add point");
                Accepted = true;
            }

            else
                Accepted = false;

            App.CommandState.StartDynamics();
        }

        public void LocateReset()
        {
            
        }

        public void Start()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            LocateCriteria lc;

            lc = App.CommandState.CreateLocateCriteria(true);
            App.CommandState.SetLocateCriteria(lc);

            App.CommandState.SetLocateCursor();

            App.CommandState.EnableAccuSnap();

            App.ShowPrompt("Select Line");
            App.ShowCommand("Create volume noise source from line, cell or shape");

            noiseForm.AttachToToolSettings(mainAddin.Instance);
        }

        #endregion
    }
}
