﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System.Text;
using Bentley.Interop.MicroStationDGN;
using NC = NoiseComputation;
using System.Windows.Forms;

namespace ISY.CAD.Noise 
{
    
    class RoadMgr : ILocateCommandEvents
    {

        private RoadForm form;

        public Point3d[] RoadPoints;
        public int roadCount;

        public RoadMgr(Point3d[] roadpoints)
        {
            RoadPoints = roadpoints;
        }

        public RoadMgr()
        {
            form = new RoadForm();
        }

        public static Point3d[] FindRoadPoints()
        {
            Bentley.Interop.MicroStationDGN.Application App = mainAddin.Instance.GetComApp;

            List<LineElement> RoadLines = scan(App, "StøyVeg");
            if (RoadLines == null)
                return null;
            List<Point3d> tempPoints = new List<Point3d>();

            foreach (LineElement el in RoadLines)
            {
                foreach (Point3d p in el.GetVertices())
                {
                    if (!tempPoints.Contains(p))
                    tempPoints.Add(p);
                }
            }
            return tempPoints.ToArray();
        }
        
        private static List<LineElement> scan(Bentley.Interop.MicroStationDGN.Application App, string level)
        {
            List<LineElement> Lines = new List<LineElement>();

            ElementScanCriteria sc = new ElementScanCriteriaClass();
            sc.ExcludeAllTypes();
            sc.IncludeType(MsdElementType.LineString);
            sc.ExcludeAllLevels();

            Level lev = App.ActiveDesignFile.Levels.Find(level);
            if (lev == null)
                return null;
            sc.IncludeLevel(lev);

            ElementEnumerator ee = App.ActiveModelReference.Scan(sc);

            while (ee.MoveNext())
            {
                LineElement el;
                el = ee.Current.AsLineElement();

                Lines.Add(el);
            }
            return Lines;
        }


        internal static void StartCommand()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            RoadMgr command = new RoadMgr();
            App.CommandState.StartLocate(command);
            App.CommandState.StartDynamics();
        }


        private void CreateRoad(Element Element)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;
            Level lev = createAndSetLevel(App, "StøyVeg");

            LineElement el;
            Point3d[] RoadPoints;

            if (Element.Type == MsdElementType.ComplexString)
            {
                List<Point3d> complexPoints = new List<Point3d>();
                ComplexStringElement cel = Element.AsComplexStringElement();
                ElementEnumerator ee = cel.GetSubElements();
                while (ee.MoveNext())
                {
                    LineElement e = ee.Current.AsLineElement();
                    foreach (Point3d p in e.GetVertices())
                        complexPoints.Add(p);
                }
                //el = App.CreateLineElement1(null, complexPoints.ToArray());
                RoadPoints = complexPoints.ToArray();
                cel.let_Level(lev);
                cel.Rewrite();
            }
            else
            {
                el = Element.AsLineElement();

                el.let_Level(lev);
                el.Rewrite();
                RoadPoints = el.GetVertices();
            }

            
            double[,] points = RunCalc.convertPoints(RoadPoints);

            //bool isAlreadyInRoads = false;
            //int countExisting = 0;
            //if (RunCalc.roads != null)
            //    for (int j = 0; j<points.Length; j++)
            //        foreach (NC.NoiseRoad r in RunCalc.roads)
            //            for (int i = 0; i < r.Points.Length; i++)
            //                if ((points[i, 0] == r.Points[i, 0]) && (points[i, 1] == r.Points[i, 1]) && (Math.Abs(p.Z - r.Points[i,2]) < (points[i, 2] + r.Points[i,2]) / 100000))
            //                {
            //                    isAlreadyInRoads = true;
            //                    countExisting++;
            //                }
            //if (isAlreadyInRoads && (countExisting >= RoadPoints.Length))
            //{
            //    MessageBox.Show("The road already exist");
            //    return;
            //}

            float roadWidth = form.GetWidth();
            float Velocity = form.GetVelocity();
            string Traffic = form.GetTraffic();
            float YDT = form.GetYDT();
            bool isTunnel = form.IsTunnel();
            bool isBridge = form.IsBridge();
            string trafficDistributionType = form.getTrafficDistributionType();

            double[] freqSpectra = new double[27];
            double[] octaveSpectra = NoiseSourceSpecForm.freqSpectra;
            if (octaveSpectra != null)
                freqSpectra = populateFreqSpectra(octaveSpectra);

            if (Traffic == "Dumper")
                freqSpectra = populateDumperSourceSpec();
            else if (Traffic == "Heavy" || Traffic == "Normal")
                freqSpectra = null;

            if (RunCalc.roads != null)
                roadCount = RunCalc.roads.Length;
            else
                roadCount = 0;

            double[] trafficDistribution = new double[10];
            if (trafficDistributionType == "General")
                trafficDistribution = null;
            else if (trafficDistributionType == "Work")
                trafficDistribution = null;
            else if (trafficDistributionType == "UserDefined")
                trafficDistribution = NoiseTimeDistributionForm.distribution;

            NC.NoiseRoad newRoad;
            //if (octaveSpectra != null)
            newRoad = new NC.NoiseRoad(points, roadWidth, Traffic, Velocity, YDT, isTunnel, isBridge, trafficDistributionType, freqSpectra, trafficDistribution);
            //else
            //    newRoad = new NC.NoiseRoad(points, roadWidth, Traffic, Velocity, YDT, isTunnel, isBridge, getTrafficDistributionType);

            MessageBox.Show("A road with "+RoadPoints.Length+" vertices created");

            RunCalc.AddRoad(newRoad);
        }

        private double[] populateDumperSourceSpec()
        {
            double[] octaveSpectra = new double[9];
            octaveSpectra[0] = -200;
            octaveSpectra[1] = 75.6f;
            octaveSpectra[2] = 92.8;
            octaveSpectra[3] = 99.4;
            octaveSpectra[4] = 100.8;
            octaveSpectra[5] = 101.6;
            octaveSpectra[6] = 100.3;
            octaveSpectra[7] = 94.1;
            octaveSpectra[8] = 83.3;

            return populateFreqSpectra(octaveSpectra);
        }


        private double[] populateFreqSpectra(double[] p)
        {
            // converting from whole octaves to 1/3 of an octave
            double third = (double)1 / 3;
            double reduceBy = 10 * Math.Log10(third);
            double[] freqSpectra = new double[27];
            
                for (int i = 0; i < 9; i++)
                    for (int j = 0; j < 3; j++)
                        freqSpectra[(3*i)+j] = Math.Pow(10,((p[i] + reduceBy)/10));
                

            return freqSpectra;
        }

        private Level createAndSetLevel(Bentley.Interop.MicroStationDGN.Application App, string p)
        {
            Level lev;
            if ((lev = App.ActiveDesignFile.Levels.Find(p)) == null)
            {
                App.ActiveDesignFile.AddNewLevel(p);
                App.ActiveDesignFile.RewriteLevels();
                lev = App.ActiveDesignFile.Levels.Find(p);
            }
            return lev;
        }

        #region ILocateCommandEvents Members

        public void Accept(Element Element, ref Point3d Point, Bentley.Interop.MicroStationDGN.View View)
        {
            CreateRoad(Element);
        }

        public void Cleanup()
        {
            form.DetachFromMicroStation();
        }

        public void Dynamics(ref Point3d Point, Bentley.Interop.MicroStationDGN.View View, MsdDrawingMode DrawMode)
        {
            
        }

        public void LocateFailed()
        {
            
        }

        public void LocateFilter(Element Element, ref Point3d Point, ref bool Accepted)
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            if (Element.Type == MsdElementType.LineString || Element.Type == MsdElementType.Line)
            {
                App.ShowPrompt("Create base line for road");
                Accepted = true;
            }
                else if (Element.Type == MsdElementType.ComplexString)
            {
                App.ShowPrompt("Create base line for road");
                Accepted = true;
            }
            else
                Accepted = false;

            App.CommandState.StartDynamics();
        }

        public void LocateReset()
        {
            
        }

        public void Start()
        {
            Bentley.Interop.MicroStationDGN.Application App = Bentley.MicroStation.InteropServices.Utilities.ComApp;

            LocateCriteria lc;

            lc = App.CommandState.CreateLocateCriteria(true);
            App.CommandState.SetLocateCriteria(lc);

            App.CommandState.SetLocateCursor();

            App.CommandState.EnableAccuSnap();

            App.ShowPrompt("Select Line");
            App.ShowCommand("Create road noise source from line");

            form.AttachToToolSettings(mainAddin.Instance);
        }

        #endregion
    }
}
